use serde::{Deserialize, Serialize};
use surrealdb::{
    engine::remote::ws::{Client, Ws},
    opt::auth::Root,
    sql::Thing,
    Response, Surreal,
};
use tracing::{debug, error, info, instrument, trace};
use url::Url;

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Website {
    /// The url that this data is found at
    site: Url,
    /// Wether or not this link has been crawled yet
    crawled: bool,
    #[serde(skip_serializing)]
    id: Option<Thing>,
}

impl Website {
    /// Creates a blank site (assumes that url param is site's root)
    pub fn new(url: &str, crawled: bool) -> Self {
        let site = match Url::parse(url) {
            Ok(a) => a,
            Err(_) => todo!(),
        };
        Self {
            id: None,
            crawled,
            site,
        }
    }

    pub fn set_crawled(&mut self) {
        trace!("Set crawled to true");
        self.crawled = true
    }

    pub fn crawled(&mut self) -> &mut bool {
        &mut self.crawled
    }

    pub fn mut_url(&mut self) -> &mut Url {
        &mut self.site
    }

    pub fn domain_str(&self) -> &str {
        self.site.as_str()
    }

    #[instrument(skip_all)]
    pub async fn links_to(&self, other: &Self, db: &Surreal<Client>) {
        let from = self.site.to_string();
        let to = other.site.to_string();
        trace!("Linking {from} to {to}");

        match db
            .query("RELATE (SELECT id FROM website WHERE site = $in) -> links_to -> (SELECT id FROM website WHERE site = $out)")
            .bind(("in", from))
            .bind(("out", to))
            .await
        {
            Ok(e) => {
                let _: Response = e;
            },
            Err(e) => {
                error!("{}", e.to_string());
            },
        }
    }

    #[instrument(skip_all)]
    pub async fn store(&mut self, db: &Surreal<Client>) {
        // check if it's been gone thru before
        let mut response = db
            .query("SELECT * FROM ONLY website WHERE site = $site LIMIT 1")
            .bind(("site", self.site.to_string()))
            .await
            .unwrap();

        if let Some(old) = response.take::<Option<Website>>(0).unwrap() {
            // site exists already
            if let Some(id) = old.id {
                let mut new = self.clone();
                new.crawled = old.crawled | new.crawled;

                match db.upsert((id.tb, id.id.to_string())).content(new).await {
                    Ok(e) => {
                        if let Some(a) = &e {
                            let _: &Record = a;
                        }
                    }
                    Err(e) => {
                        error!("{}", e);
                    }
                };
            }
        } else {
            // sites hasn't existed yet
            match db.create("website").content(self.clone()).await {
                Ok(e) => {
                    let _: Option<Record> = e;
                    if let Some(a) = &e {
                        let _: &Record = a;
                    }
                }
                Err(a) => error!("{:?}", a),
            };
        }
    }
}

impl ToString for Website {
    fn to_string(&self) -> String {
        self.site.to_string()
    }
}

#[derive(Debug, Serialize)]
pub struct Email {
    pub email: String,
}

#[derive(Debug, Deserialize)]
pub struct Record {
    #[allow(dead_code)]
    pub id: Thing,
}

pub async fn connect() -> surrealdb::Result<Surreal<Client>> {
    // Connect to the server
    let db = Surreal::new::<Ws>("127.0.0.1:8000").await?;

    // Signin as a namespace, database, or root user
    db.signin(Root {
        username: "root",
        password: "root",
    })
    .await?;

    // Select a specific namespace / database
    db.use_ns("test").use_db("time").await?;

    Ok(db)
}
